#ifndef __BRISLAWN__
#define __BRISLAWN__

#include "wavelet.h"

invertlowFunc invertlow_c;
inverthighFunc inverthigh_c;
transformFunc transform_c;

extern "C" { // moved here from wavelet.h in v.0.30 by Fizick
extern transformFunc transform_3dne, transform_sse;
extern invertlowFunc invertlow_3dne, invertlow_sse;
extern inverthighFunc inverthigh_3dne, inverthigh_sse;
}

class Brislawn : public Wavelet {
public:
	Brislawn(int x, int y, int n);
	_inline void transform2d (float *input, int hsize, int vsize, int nsteps);
	_inline void invert2d (float *input, int hsize, int vsize, int nsteps);
	_inline void invert_step(float register* input, float register* output,
							 int size);
	_inline void transform_step(float register* input, float register* output,
								int size, int lowSize);
	_inline void symmetric_extension(float *output, int size, int right_ext);
	_inline void antisymmetric_extension(float *output, int size, int right_ext);
};

class Brislawn_3DNowExt : public Brislawn
{
public:
	Brislawn_3DNowExt(int x, int y, int n) : Brislawn(x ,y, n)
	{
		Transform = transform_3dne;
		InvertLow = invertlow_3dne;
		InvertHigh = inverthigh_3dne; // it is more fast
	};
};

class Brislawn_3DNow : public Brislawn
{
public:
	Brislawn_3DNow(int x, int y, int n) : Brislawn(x ,y, n)
	{
		Transform = transform_3dne;
		InvertLow = invertlow_3dne;
		InvertHigh = inverthigh_3dne;
	};
};

class Brislawn_SSE : public Brislawn
{
public:
	Brislawn_SSE(int x, int y, int n) : Brislawn(x ,y, n)
	{
		Transform = transform_sse;
//		InvertLow = invertlow_sse;
//		InvertHigh = inverthigh_sse;
		InvertLow = invertlow_c; // It seems, invert C versions is faster - v0.35
		InvertHigh = inverthigh_c; 
	};
};

class Brislawn_C : public Brislawn
{
public:
	Brislawn_C(int x, int y, int n) : Brislawn(x ,y, n)
	{
		Transform = transform_c;
		InvertLow = invertlow_c;
		InvertHigh = inverthigh_c;
	};
};

#endif