#include "VBL610.h"

//  J. D. Villasenor, B. Belzer, and J. Liao, (6,10)  wavelet  functions  - added in v.0.32 by Fizick

/*
  VagueDenoiser plugin for Avisynth -- a wavelet denoiser plugin
  VagueDenoiser (C) 2003, LeFunGus
  VagueDenoiser assembly parts, debug and adaptative thresholding (C) Kurosu
  WienerChop mode, 9/7 and 6/10 wavelets (c) 2004 Fizick (bag@hotmail.ru)

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details. 
*/

#define NPAD 10


VBL610::VBL610(int x, int y, int n) : Wavelet(x ,y, n)
{
	wltnumber = VBL6_10;
	temp_in = (float*) _mm_malloc((32+max(x,y))*sizeof(float),128);
	temp_out = (float*) _mm_malloc((32+max(x,y))*sizeof(float),128);
	temp2 = (float*) _mm_malloc((32+max(x,y))*sizeof(float),128);
}

#define OFFSET		6		//better alignment? 10+6=16 -> cacheline

void VBL610::transform2d(float *input, int hsize, int vsize, int nsteps)
{
	int j;
	int hLowSize = hsize;
	int vLowSize = vsize;

	while (nsteps--)
	{
		int lowSize=(hLowSize+1)>>1;
		float *inputp = input;
		for (j = 0; j < vLowSize; j++) 
		{
			copy (inputp, temp_in+NPAD+OFFSET, hLowSize); // fast BitBlt block copy, aligned src and dst
			transform_step (temp_in+OFFSET, temp_out, hLowSize, lowSize);
			//symmetric_extension(temp_in+OFFSET, hLowSize, 2);
			//Transform(temp_in+OFFSET, temp_out, lowSize);
			copy (temp_out+NPAD, inputp, hLowSize);  // fast BitBlt block copy, but not-aligned nor src nor dst
			inputp += hsize;
		}

		lowSize=(vLowSize+1)>>1;
		inputp = input;
			for (j = 0; j < hLowSize; j++)
			{
				copy (inputp, hsize, temp_in+NPAD+OFFSET, vLowSize); // slow for size mod64
				transform_step(temp_in+OFFSET,temp_out,vLowSize,lowSize);
				//symmetric_extension(temp_in+OFFSET, vLowSize, 2);
				//Transform(temp_in+OFFSET, temp_out, lowSize);
				copy (temp_out+NPAD,inputp,hsize,vLowSize);  // the slowest for size mod64 (and general too)
				inputp++;
			}
		hLowSize = (hLowSize+1)>>1;
		vLowSize = (vLowSize+1)>>1;
	}
}

void VBL610::transform_step (float* input, float* output,
									int size, int lowSize)
{
	//	symmetric_extension(input, size, 2); // Brislawn 10/10
	symmetric_extension(input, size, 2, 2); // even filter length
//	symmetric_extension(input, size, 1, 1); // Brislawn 9/7
	Transform(input, output, lowSize);
}

void VBL610::invert2d(float *input, int hsize, int vsize, int nsteps)
{
	int i;

	hLowSize[0] = (hsize+1)>>1;
	hHighSize[0] = hsize>>1;
	vLowSize[0] = (vsize+1)>>1;
	vHighSize[0] = vsize>>1;
	
	for (i = 1; i < nsteps; i++)
	{
		hLowSize[i] = (hLowSize[i-1]+1)>>1;
		hHighSize[i] = hLowSize[i-1]>>1;
		vLowSize[i] = (vLowSize[i-1]+1)>>1;
		vHighSize[i] = vLowSize[i-1]>>1;
	}

	while (nsteps--)
	{
		int idx = vLowSize[nsteps]+vHighSize[nsteps];
		int idx2= hLowSize[nsteps]+hHighSize[nsteps];
			for (i=0;i<idx2;i++)
			{
				float* idx3=input+i;
				copy (idx3,hsize,temp_in+NPAD,idx); // slow for size mod64, not-aligned src, not-aligned dst,
				invert_step (temp_in,temp_out,idx);
				copy (temp_out+NPAD,idx3,hsize,idx); // slowest for size mod64 (and general too)
			}

		float* idx3=input;
		for (i=0;i<idx;i++)
		{
			copy(idx3,temp_in+NPAD+OFFSET,idx2); // fast BitBlt block copy, aligned dest (and src if hsize=mod16)
			invert_step(temp_in+OFFSET,temp_out,idx2);
			copy(temp_out+NPAD,idx3,idx2); // fast BitBlt block copy, not-aligned src, ( aligned dest if hsize=mod16)
			idx3 += hsize;
		}
	}
}

void VBL610::invert_step(float* input, float* output, int size)
{
	int findex, left_ext, right_ext;
	// amount of low and high pass -- if odd # of values, extra will be
	//   low pass
	int lowSize = (size+1)>>1, highSize = size>>1;
	float *temp = temp2;
	
	DLLEnv->BitBlt( (BYTE*)temp+NPAD*sizeof(float),0,
					(const BYTE*)input+NPAD*sizeof(float),0,
					lowSize*sizeof(float),1);

//	left_ext = 1; // for odd filter
	left_ext = 2; // for even filter
	right_ext = (size % 2 == 0) ? 2 : 1; // 
	symmetric_extension (temp, lowSize, left_ext, right_ext);
	
	// coarse  detail
	// HHHHHHHHGGGGGGGG --> xxxxxxxxxxxxxxxx
	mem_set(output,0,(NPAD+NPAD+size)*sizeof(float));
//	findex=(size+3)>>1; // was for even filter  Brislawn 10/10
	findex=(size+3)>>1; // was for even filter  Brislawn 10/10
//	findex=(size+2)>>1;  // was for odd filter 9,7 ,  =(size-1-SLOWFIRST)/2
	InvertLow(output, temp, findex);

	DLLEnv->BitBlt( (BYTE*)temp+NPAD*sizeof(float),0,
					(const BYTE*)input+(NPAD+lowSize)*sizeof(float),0,
					highSize*sizeof(float),1);

	left_ext = 2;
	
//	right_ext = (size % 2 == 0) ? 1 : 2; // CDF97  9,7 filter is odd, 
	right_ext = (size % 2 == 0) ? 2 : 1; // VBL610  6/10 filter is even, 
	// VBL610  6/10 filter is even, then asymmetric extension
	antisymmetric_extension (temp, highSize, left_ext, right_ext);
	InvertHigh(output, temp, findex);
}

// Do symmetric extension of data using prescribed symmetries
//   Original values are in output[npad] through output[npad+size-1]
//   New values will be placed in output[0] through output[npad] and in
//      output[npad+size] through output[2*npad+size-1] (note: end values may
//      not be filled in)
//    extension at left bdry is ... 3 2 1 0 | 0 1 2 3 ...
//    same for right boundary
//    if right_ext=1 then ... 3 2 1 0 | 1 2 3
//
//   symmetry = 1  -> extend symmetrically
//   symmetry = -1 -> extend antisymmetrically
void VBL610::symmetric_extension (float *output, int size, int left_ext, int right_ext)
{
	// VBL610 Filter is even 
	int i,idx,first,last,originalLast,originalSize,period,nextend;
	first = NPAD;// was 9 ?;
	last = NPAD-1+size; 
	
	originalLast=last;
	originalSize=originalLast-NPAD+1;
	period = ((last-NPAD+1)<<1)-(right_ext==1) - (left_ext == 1) ;
	
//	output[9] = output[10];
	if (left_ext == 2)
		output[--first] = output[NPAD];
	if(right_ext==2)
		output[++last] = output[originalLast];
	
	// extend left end
	nextend = first;
	for (i = 0; i < nextend; i++) 
		output[--first] = output[NPAD+1+i];
	
	idx=NPAD+NPAD-1+size;
	
	// extend right end
	nextend=idx-last;
	for(i=0;i<nextend;i++) 
		output[++last]=output[originalLast-1-i];

	// should have full period now -- extend periodically
//  while (last < 2*NPAD+size-1) {
//    last++;
 //   output[last] = output[last-period];
 // }
}

void VBL610::antisymmetric_extension (float *output, int size, int left_ext, int right_ext)
{
	// VBL610 6,10 Filter is even
	// (it is from Brislawn 10/10).

	int i,idx,first,last,originalLast,originalSize,period,nextend;
	first=NPAD;
	last=NPAD-1+size;
	
	// left_ext=2 always for even filters - not check
//	if (left_ext == 1)	output[--first] = 0;
	if (right_ext==1) output[++last] = 0;
	
	originalLast=last;
	originalSize=originalLast-NPAD+1;
	period = ((last-NPAD+1)<<1)-(right_ext==1); // -(left_ext == 1)
	
	output[--first] = -output[NPAD];
	if(right_ext==2)
		output[++last] = -output[originalLast];
	
	// extend left end
	for (i = 0; i < NPAD-1; i++) 
		output[--first] = -output[NPAD+1+i];
 
	// should have full period now -- extend periodically
//  while (first > 0) {
 //   first--;
//    output[first] = output[first+period];
//  }
	
	idx=NPAD+NPAD-1+size;
	
	// extend right end
	nextend=idx-last;
	for(i=0;i<nextend;i++) 
		output[++last] = -output[originalLast-1-i];

  // should have full period now -- extend periodically
//  while (last < 2*NPAD+size-1) {
 //   last++;
 //   output[last] = output[last-period];
 // }

}


__declspec(align(32)) float analysis3Low[6] = {
  -1.290777652578771e-01f,   4.769893003875977e-02f,   7.884856164056651e-01f,
   7.884856164056651e-01f,   4.769893003875977e-02f,  -1.290777652578771e-01f };
__declspec(align(32)) float analysis3High[10] = {
   1.891422775349768e-02f, -6.989495243807747e-03f, -6.723693471890128e-02f,
  -1.333892255971154e-01f,  6.150507673110278e-01f, -6.150507673110278e-01f,
   1.333892255971154e-01f,  6.723693471890128e-02f,  6.989495243807747e-03f,
  -1.891422775349768e-02f };
__declspec(align(32)) float synthesis3Low[10] = {
  1.891422775349768e-02f,  6.989495243807747e-03f, -6.723693471890128e-02f,
  1.333892255971154e-01f,  6.150507673110278e-01f,  6.150507673110278e-01f,
  1.333892255971154e-01f, -6.723693471890128e-02f,  6.989495243807747e-03f,
  1.891422775349768e-02f };
__declspec(align(32)) float synthesis3High[6] = {
  -1.290777652578771e-01f,  -4.769893003875977e-02f,   7.884856164056651e-01f,
  -7.884856164056651e-01f,   4.769893003875977e-02f,   1.290777652578771e-01f };

	// Brislawn2 (10/10) sizes
//#define ALOWSIZE 10 
//#define AHIGHSIZE 10 
//#define SLOWSIZE 10 
//#define SHIGHSIZE 10 
// Brislawn2 first indexes
//#define ALOWFIRST -4
//#define AHIGHFIRST -4
//#define SLOWFIRST -4
//#define SHIGHFIRST -4

	// VBL610 sizes
//#define ALOWSIZE1 6 
//#define AHIGHSIZE1 10 
//#define SLOWSIZE1 10 
//#define SHIGHSIZE1 6 
// VBL610 first indexes
//#define ALOWFIRST1 -2
//#define AHIGHFIRST1 -4
//#define SLOWFIRST1 -4
//#define SHIGHFIRST1 -2


void transform3_c(float* input, float* output, const unsigned int lowSize)
{
	for (unsigned int i=NPAD;i<NPAD+lowSize;i++)
	{
		float a,b,c,d,e,f;
//		a=input[2*i-NPAD+ALOWFIRST1]*analysis1Low[0]; // first (use symmetry)
		a=input[2*i-12]*analysis3Low[0]; 
		b=input[2*i-11]*analysis3Low[1]; 
		c=input[2*i-10]*analysis3Low[2]; 
		d=input[2*i-9]*analysis3Low[2]; //3
		e=input[2*i-8]*analysis3Low[1]; //4
		f=input[2*i-7]*analysis3Low[0]; //5
//		k=input[2*i-NPAD+ALOWSIZE1-1+ALOWFIRST1]*analysis1Low[ALOWSIZE1-1]; // last
		output[i]=a+b+c+d+e+f;
	}
	for (i=NPAD;i<NPAD+lowSize;i++)
	{
		float a,b,c,d,e,f,g,h,k,l;
//		a=input[2*i-NPAD+AHIGHFIRST1]*analysis1High[0]; // first (use symmetry)
		a=input[2*i-14]*analysis3High[0];
		b=input[2*i-13]*analysis3High[1];
		c=input[2*i-12]*analysis3High[2];
		d=input[2*i-11]*analysis3High[3];
		e=input[2*i-10]*analysis3High[4];
		f=input[2*i-9]*analysis3High[4]; //5
		g=input[2*i-8]*analysis3High[3];//6
		h=input[2*i-7]*analysis3High[2];//7
		k=input[2*i-6]*analysis3High[1];//8
		l=input[2*i-5]*analysis3High[0];//9
//		g=input[2*i-NPAD+AHIGHSIZE-1+AHIGHFIRST1]*analysis1High[AHIGHSIZE1-1]; // last
		output[i+lowSize]=a+b+c+d+e-f-g-h-k-l;
	}
};

void invertlow3_c(float* output, float* temp, const unsigned int findex)
{
	float a,b,c,d,e;
	// lindex1 = -(SLOWSIZE1-1+SLOWFIRST1)/2
	// lindex = -2 (Brislaw2)
	// NPAD+lindex = 8
	// lindex1 = -1 (CDF97)
	// NPAD+lindex1 = 9
	// findex here is lowSize here
	// NPAD+findex+1 is last
	for(unsigned int i=8;i<findex+12;i++)
	{
		a=temp[i]*synthesis3Low[0];
		b=temp[i]*synthesis3Low[1];
		c=temp[i]*synthesis3Low[2];
		d=temp[i]*synthesis3Low[3];
		e=temp[i]*synthesis3Low[4];
		// for rest use symmetry 
		

//		output[2*i-NPAD+SLOWFIRST1]+=a;
		output[2*i-14]+=a;
		output[2*i-13]+=b;
		output[2*i-12]+=c;
		output[2*i-11]+=d;
		output[2*i-10]+=e;
		output[2*i-9] +=e;
		output[2*i-8] +=d;
		output[2*i-7] +=c;
		output[2*i-6] +=b;
		output[2*i-5] +=a;
	}
};

void inverthigh3_c(float* output, float* temp, const unsigned int findex)
{
	// lindex1 = -(SHIGHSIZE1-1+SHIGHFIRST1)/2
	// lindex = -2 (Brislawn1010)
	// NPAD+lindex = 8
	// lindex1 = -2 (CDF97)
	// NPAD+lindex1 = 8
	// findex here is lowSize here
	// NPAD+findex+1 is last
	// lindex = -1 (VBL610)
	// NPAD+lindex1 = 9
//	for(unsigned int i=8;i<lowSize+12;i++)
	for (unsigned int i=9;i<findex+12;i++)
	{
		float a,b,c;
		
		a=temp[i]*synthesis3High[0];
		b=temp[i]*synthesis3High[1];
		c=temp[i]*synthesis3High[2];
		
//		output[2*i-NPAD+SHIGHFIRST1]+=a;
		output[2*i-12]+=a;
		output[2*i-11]+=b;
		output[2*i-10]+=c;
		output[2*i-9] -=c;
		output[2*i-8] -=b;
		output[2*i-7] -=a;
	}
};