//  VagueDenoiser plugin for Avisynth -- a wavelet denoiser plugin
//  VagueDenoiser (C) 2003, LeFunGus
//  VagueDenoiser assembly parts, debug and adaptative thresholding (C) Kurosu
//  WienerChop mode, 9/7 and 6/10 wavelets, AutoThreshold (c)2004 Fizick (bag@hotmail.ru)
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation.
//  
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//  
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
//  Include file
//  Version 0.35, September 16, 2005 
  
#ifndef		__VD__
#define		__VD__

#include <windows.h>
#include <malloc.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include "avisynth.h"
#include "brislawn.h"
#include "CDF97.h" // v0.30 by Fizick
#include "VBL610.h" // v0.32 by Fizick

void InitTable(int width, int height, int nsteps, unsigned char* LUT);
void MultiScaleThresholding(float *block, int length, float thresh, int width, int height, int nsteps, float ratio, float percent, unsigned char *LUT);


extern "C" {
/***** ByteToFloat ****/
typedef void(B2FFunc) (const unsigned int h,
					   const unsigned int w,
					   const unsigned char *src,
					   const unsigned int srcPitch,
					   float *blockptr,
					   const unsigned int blockPitch);//  blockPitch added in v.0.33
typedef B2FFunc *B2FFuncPtr;
extern B2FFunc b2f_3dne, b2f_sse;

/***** ByteToFloat 1D ****/
typedef void(B2F1DFunc) (const unsigned int w,
					     const unsigned char *src,
					     float *blockptr);
typedef B2F1DFunc *B2F1DFuncPtr;
extern B2F1DFunc b2f1D_3dne, b2f1D_sse;

/***** FloatToByte ****/
typedef void(F2BFunc) (const unsigned int h,
					   const unsigned int w,
					   unsigned char *dst,
					   const unsigned int srcPitch,
					   float *blockptr,
					   const unsigned int blockPitch);//  blockPitch added in v.0.33
typedef F2BFunc *F2BFuncPtr;
extern F2BFunc f2b_3dne, f2b_sse;

/***** ByteToFloat 1D ****/
typedef void(F2B1DFunc) (const unsigned int w,
					     unsigned char *dst,
					     float *blockptr);
typedef F2B1DFunc *F2B1DFuncPtr;
extern F2B1DFunc f2b1D_3dne, f2b1D_sse;

/***** Filter *****/
typedef void(FilterFunc)(float* block, const unsigned int b, const float threshold,
						 const unsigned int width, const unsigned int height, const unsigned int nsteps, float percent); //v0.35
typedef FilterFunc *FilterFuncPtr;
extern FilterFunc ht_3dne, st_3dne, qt_3dne;
extern FilterFunc ht_sse, st_sse, qt_sse, smt_sse;

// WienerChop
typedef void (WienerChopFunc)(const float * block1, float *block, const unsigned int length, float sigmaNoise); // 0.30-0.32
extern WienerChopFunc wiener_chop_3dne; // function is ASM coded by Kurosu, added in v.0.32 
typedef WienerChopFunc *WienerChopFuncPtr; 

// AutoThreshold
typedef void(AutoThresholdFunc)(const float *block, int w, int h, int nsteps,  float wratio, float * threshold);// added in v.0.30-0.32
extern AutoThresholdFunc auto_th_3dne; // function is ASM coded by Kurosu, added in v.0.32 
typedef AutoThresholdFunc *AutoThresholdFuncPtr;

}

B2FFunc b2f_c;
B2F1DFunc b2f1D_c;
F2BFunc f2b_c;
F2B1DFunc f2b1D_c;
FilterFunc ht_c, st_c, qt_c, smt_c; // v0.35

WienerChopFunc wiener_chop_c; // function is added in v.0.30, corrected in 0.31-0.32 by Fizick
AutoThresholdFunc auto_th_c; // function is added in v.0.30, corrected in 0.31-0.32 by Fizick


#define WIDTHPAD 4 // pad of the width (fixed) for better cache - added in v.0.33, replaced from 2 (some problems with SSE) to 4 in v.0.33.1 

void mirror_pad(float *blockptr, const unsigned int h, const unsigned int w); // added in v.0.33

void partial_denoise( unsigned char* dst, int dstPitch, const unsigned char* src, int srcPitch,
			   unsigned int w, unsigned int h, float percentDst);

class VagueDenoiser : public GenericVideoFilter
{   
public:
	float *block;
	float *block1;// v.0.30 Fizick
	unsigned char *buffY, *buffU, *buffV;
	unsigned char *auxbuffY, *auxbuffU, *auxbuffV; // added in v.0.34
	unsigned char *buffer;
//	unsigned char *LUT; // removed as unused in 0.32
//	float *maxs;   // removed as unused in 0.32
	unsigned char *LUTY; // added in v0.35
	unsigned char *LUTU; // added in v0.35

	float thresh;
	int method;
	int nsteps; // may be changed (limited by max) in v.0.33.6
	float lumaT; //  v.0.30 by Fizick
	float chromaT; //  v.0.30 by Fizick
	bool debug;
	bool itlc;
	int wlt; //  v.0.30 by Fizick
	bool Wiener; // v.0.30 by Fizick
	float wratio; // v.0.31 by Fizick
	__declspec(align(16)) Wavelet *wavelet;
	__declspec(align(16)) Wavelet *wavelet1;// v.0.30 Fizick
	float percent ; // v.0.33 v0.35 Fizick
	PClip auxClip ; // v.0.34 Fizick
	bool useAuxClip; // v.0.34 Fizick
	IScriptEnvironment *VDEnv;
	
	int width, height;
	int widthU, heightU;
	int bufSize, bufSizeU; // renamed in place of SizeEven in 0.34.2
	int pitch, pitchU; // added in v.0.34.2 Fizick
	char debugbuf[80]; // v.0.33 Fizick

	VagueDenoiser(PClip _child1, double _lumaT, const int _method,
				  const int _nsteps, double _chromaT, bool _dbg, bool _itlc,
				  int _wlt,  bool _Wiener, float _wratio, float _percent, PClip _auxClip,
				  IScriptEnvironment* env);
	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
	~VagueDenoiser();

	void filterBlockYUY2(const unsigned char* src, unsigned char* dst,
						 int srcPitch, int dstPitch,
						 const unsigned char* aux, int auxPitch);// v.0.34
	void filterBlockRGB(const unsigned char* src, unsigned char* dst,
						int srcPitch, int dstPitch, char numb,
						const unsigned char* aux, int auxPitch);
	void filterBlockPlane(int h, int w, const unsigned char* src,
						  unsigned char* dst, int srcPitch, int dstPitch,
						  const unsigned char* aux, int auxPitch); // v.0.34

	__declspec(align(16)) FilterFuncPtr filter;
	__declspec(align(16)) B2FFuncPtr byte2float;
	__declspec(align(16)) F2BFuncPtr float2byte;
	__declspec(align(16)) B2F1DFuncPtr byte2float1D;
	__declspec(align(16)) F2B1DFuncPtr float2byte1D;

	__declspec(align(16)) WienerChopFuncPtr WienerChop; // added in 0.32 here
	__declspec(align(16)) AutoThresholdFuncPtr AutoThreshold; // added in 0.32 here
};

class VagueDenoiser3DNowExt : public VagueDenoiser
{
public:
	VagueDenoiser3DNowExt(PClip _child1, double _lumaT, const int _method,
					   const int _nsteps, double _chromaT, bool _dbg, bool _itlc,
						int _wlt, bool _Wiener, float _wratio, float _percent, PClip _auxClip,
					   IScriptEnvironment* env) :
    VagueDenoiser(_child1,  _lumaT, _method, _nsteps, _chromaT, _dbg, _itlc, _wlt, _Wiener, _wratio, _percent, _auxClip, env)
	{
		OutputDebugString("VagueDenoiser: using SSE / 3DNow!Extended code");
		switch (_wlt)
		{
		case BRISLAWN:
			wavelet = new Brislawn_3DNowExt(vi.width, vi.height, nsteps); // replace _nsteps to  nsteps (variable) here and below in v 0.33.6
			if (_Wiener) wavelet1 = new CDF97_3DNow(vi.width, vi.height, nsteps);
			break;
		case CDF9_7:
			wavelet = new CDF97_3DNowExt(vi.width, vi.height, nsteps);
			if (_Wiener) wavelet1 = new VBL610_3DNow(vi.width, vi.height, nsteps);
			break;
		case VBL6_10:
			wavelet = new VBL610_3DNowExt(vi.width, vi.height, nsteps);
			if (_Wiener) wavelet1 = new CDF97_3DNow(vi.width, vi.height, nsteps);
			break;
		default:
			env->ThrowError("VagueDenoiser: Wavelet N%i doesn't exist", _wlt);
		}
		wavelet->DLLEnv = env;
		if (_Wiener) wavelet1->DLLEnv = env;

		switch(method)
		{
		case 0: //Hard thresholding
			filter = ht_sse; // v0.35
			break;
		case 1: //Soft thresholding
			filter = st_c;// new assembler version is not implemented yet - use C v0.35
			break;
		case -1: //No thresholding - debug
			break;
		case 2:
			//env->ThrowError("VagueDenoiser: This mode isn't available anymore"); // re-enabled multilevel in v0.35
			break;
		case 3://Qian thresholding
			filter = qt_sse; // v0.35
			break;
		case 4: //Smooth thresholding - added in v0.35
			filter = smt_sse;
			break;
		default:
			env->ThrowError("VagueDenoiser: Unvalid filtering mode");
		}
		byte2float = b2f_3dne;
		float2byte = f2b_3dne;
		byte2float1D = b2f1D_3dne;
		float2byte1D = f2b1D_3dne;
		WienerChop = wiener_chop_3dne; // added in 0.32 here
		AutoThreshold = auto_th_3dne; // added in 0.32 here
	}
};

class VagueDenoiser3DNow : public VagueDenoiser
{
public:
	VagueDenoiser3DNow(PClip _child1, double _lumaT, const int _method,
					   const int _nsteps, double _chromaT, bool _dbg, bool _itlc,
						int _wlt, bool _Wiener, float _wratio, float _percent, PClip _auxClip,
					   IScriptEnvironment* env) :
    VagueDenoiser(_child1, _lumaT, _method, _nsteps, _chromaT, _dbg, _itlc, _wlt, _Wiener, _wratio, _percent,  _auxClip, env)
	{
		OutputDebugString("VagueDenoiser: using 3DNow! code");
		switch (_wlt)
		{
		case BRISLAWN:
			wavelet = new Brislawn_3DNow(vi.width, vi.height, nsteps);
			if (_Wiener) wavelet1 = new CDF97_3DNow(vi.width, vi.height, nsteps);
			break;
		case CDF9_7:
			wavelet = new CDF97_3DNow(vi.width, vi.height, nsteps);
			if (_Wiener) wavelet1= new VBL610_3DNow(vi.width, vi.height, nsteps);
			break;
		case VBL6_10:
			wavelet = new VBL610_3DNow(vi.width, vi.height, nsteps);
			if (_Wiener) wavelet1= new CDF97_3DNow(vi.width, vi.height, nsteps);
			break;
		default:
			env->ThrowError("VagueDenoiser: Wavelet N%i doesn't exist", _wlt);
		}
		wavelet->DLLEnv = env;
		if (_Wiener) wavelet1->DLLEnv = env;

		switch(method)
		{
		case 0: //Hard thresholding
			filter = ht_c; // new assembler version is not implemented yet - use C v0.35
			break;
		case 1: //Soft thresholding
			filter = st_c;// new assembler version is not implemented yet - use C v0.35
			break;
		case -1: //No thresholding - debug
			break;
		case 2:
			//env->ThrowError("VagueDenoiser: This mode isn't available anymore");// re-enabled multilevel in v0.35
			break;
		case 3://Qian thresholding
			filter = qt_c;  // new assembler version is not implemented yet - use C v0.35
			break;
		case 4: //Smooth thresholding
			filter = smt_c;// new assembler version is not implemented yet - use C v0.35
			break;
		default:
			env->ThrowError("VagueDenoiser: Unvalid filtering mode");
		}
		byte2float = b2f_3dne;
		float2byte = f2b_3dne;
		byte2float1D = b2f1D_3dne;
		float2byte1D = f2b1D_3dne;
		WienerChop = wiener_chop_3dne; // added in 0.32 here
		AutoThreshold = auto_th_3dne; // added in 0.32 here
	}
};

class VagueDenoiserSSE : public VagueDenoiser
{
public:
	VagueDenoiserSSE(PClip _child1, double _lumaT, const int _method,
					   const int _nsteps, double _chromaT, bool _dbg, bool _itlc,
						int _wlt, bool _Wiener, float _wratio, float _percent, PClip _auxClip,
						IScriptEnvironment* env) :
    VagueDenoiser(_child1, _lumaT, _method, _nsteps, _chromaT, _dbg, _itlc, _wlt, _Wiener, _wratio, _percent, _auxClip, env)
	{
		OutputDebugString("VagueDenoiser: using SSE code");
		switch (_wlt)
		{
		case BRISLAWN:
			wavelet = new Brislawn_SSE(vi.width, vi.height, nsteps);
			if (_Wiener) wavelet1 = new CDF97_SSE(vi.width, vi.height, nsteps);
			break;
		case CDF9_7:
			wavelet = new CDF97_SSE(vi.width, vi.height, nsteps);
			if (_Wiener) wavelet1= new VBL610_SSE(vi.width, vi.height, nsteps);
			break;
		case VBL6_10:
			wavelet = new VBL610_SSE(vi.width, vi.height, nsteps);
			if (_Wiener) wavelet1= new CDF97_SSE(vi.width, vi.height, nsteps);
			break;
		default:
			env->ThrowError("VagueDenoiser: Wavelet N%i doesn't exist", _wlt);
		}
		wavelet->DLLEnv = env;
		if (_Wiener) wavelet1->DLLEnv = env;

		switch(method)
		{
		case 0: //Hard thresholding
			filter = ht_sse; 
			break;
		case 1: //Soft thresholding
			filter = st_c; // new assembler version is not implemented yet - use C v0.35
			break;
		case -1: //No thresholding - debug
			break;
		case 2:
			//env->ThrowError("VagueDenoiser: This mode isn't available anymore");// re-enabled multilevel in v0.35
			break;
		case 3://Qian (garrote) thresholding
			filter = qt_sse;
			break;
		case 4: //Smooth thresholding
			filter = smt_sse;// v0.35
			break;
		default:
			env->ThrowError("VagueDenoiser: Unvalid filtering mode");
		}

		// SSE convert fuctions had a some bugs from v.0.29
		// Were temporary replaced by C version in v.0.32 by Fizick, reenabled in v.0.32.1
		byte2float = b2f_sse; 
		float2byte = f2b_sse; 
		byte2float1D = b2f1D_sse; 
		float2byte1D = f2b1D_sse; 

		WienerChop = wiener_chop_c; // added in 0.32 here
		AutoThreshold = auto_th_c; // added in 0.32 here
	}
};

class VagueDenoiserC : public VagueDenoiser
{
public:
	VagueDenoiserC(PClip _child1, double _lumaT, const int _method,
					   const int _nsteps, double _chromaT, bool _dbg, bool _itlc,
						int _wlt, bool _Wiener, float _wratio, float _percent, PClip _auxClip,
					   IScriptEnvironment* env) :
    VagueDenoiser(_child1, _lumaT, _method, _nsteps, _chromaT, _dbg, _itlc,  _wlt, _Wiener, _wratio, _percent, _auxClip, env)
	{
		OutputDebugString("VagueDenoiser: using C code");
		switch (_wlt)
		{
		case BRISLAWN:
			wavelet = new Brislawn_C(vi.width, vi.height, nsteps);
			if (_Wiener) wavelet1 = new CDF97_C(vi.width, vi.height, nsteps);
			break;
		case CDF9_7:
			wavelet = new CDF97_C(vi.width, vi.height, nsteps);
			if (_Wiener) wavelet1 = new VBL610_C(vi.width, vi.height, nsteps);
			break;
		case VBL6_10:
			wavelet = new VBL610_C(vi.width, vi.height, nsteps);
			if (_Wiener) wavelet1 = new CDF97_C(vi.width, vi.height, nsteps);
			break;
		default:
			env->ThrowError("VagueDenoiser: Wavelet N%i doesn't exist", _wlt);
		}
		wavelet->DLLEnv = env;
		if (_Wiener) wavelet1->DLLEnv = env;

		switch(method)
		{
		case 0: //Hard thresholding
			filter = ht_c;
			break;
		case 1: //Soft thresholding
			filter = st_c;
			break;
		case -1: //No thresholding - debug
			break;
		case 2:
			//env->ThrowError("VagueDenoiser: This mode isn't available anymore");// re-enabled multilevel in v0.35
			break;
		case 3://Qian (garrote) thresholding
			filter = qt_c;
			break;
		case 4: //Smooth thresholding
			filter = smt_c;
			break;
		default:
			env->ThrowError("VagueDenoiser: Unvalid filtering mode");
		}
		byte2float = b2f_c;
		float2byte = f2b_c;
		byte2float1D = b2f1D_c;
		float2byte1D = f2b1D_c;
		WienerChop = wiener_chop_c; // added in 0.32 here
		AutoThreshold = auto_th_c; // added in 0.32 here
	}
};
#endif
