#ifndef __CDF97__
#define __CDF97__

// Header for CDF97 (9/7) wavelet - added in v.0.30 by Fizick

#include "wavelet.h"

invertlowFunc invertlow1_c;
inverthighFunc inverthigh1_c;
transformFunc transform1_c;

extern "C" {
extern transformFunc transform1_3dne;//, transform1_sse;
extern invertlowFunc invertlow1_3dne;//, invertlow1_sse;
extern inverthighFunc inverthigh1_3dne;//, inverthigh1_sse;
}

class CDF97 : public Wavelet {
public:
	CDF97(int x, int y, int n);
	_inline void transform2d (float *input, int hsize, int vsize, int nsteps);
	_inline void invert2d (float *input, int hsize, int vsize, int nsteps);
	_inline void invert_step(float register* input, float register* output,
							 int size);
	_inline void transform_step(float register* input, float register* output,
								int size, int lowSize);
	_inline void symmetric_extension(float *output, int size, int left_ext, int right_ext);
	_inline void antisymmetric_extension(float *output, int size, int left_ext, int right_ext);
};

class CDF97_3DNowExt : public CDF97
{
public:
	CDF97_3DNowExt(int x, int y, int n) : CDF97(x ,y, n)
	{
		Transform = transform1_3dne;
//		InvertLow = invertlow1_3dne;
//		InvertHigh = inverthigh1_3dne;
//		Transform = transform1_c;
		InvertLow = invertlow1_c; // it is more fast
		InvertHigh = inverthigh1_c; // it is more fast
	};
};

class CDF97_3DNow : public CDF97
{
public:
	CDF97_3DNow(int x, int y, int n) : CDF97(x ,y, n)
	{
		Transform = transform1_3dne;
//		InvertLow = invertlow1_3dne;
//		InvertHigh = inverthigh1_3dne;
//		Transform = transform1_c;
		InvertLow = invertlow1_c;// it is more fast
		InvertHigh = inverthigh1_c;// it is more fast
	};
};

class CDF97_SSE : public CDF97
{
public:
	CDF97_SSE(int x, int y, int n) : CDF97(x ,y, n)
	{
//		Transform = transform1_sse;
//		InvertLow = invertlow1_sse;
//		InvertHigh = inverthigh1_sse;
		Transform = transform1_c;
		InvertLow = invertlow1_c;
		InvertHigh = inverthigh1_c;
	};
};

class CDF97_C : public CDF97
{
public:
	CDF97_C(int x, int y, int n) : CDF97(x ,y, n)
	{
		Transform = transform1_c;
		InvertLow = invertlow1_c;
		InvertHigh = inverthigh1_c;
	};
};

#endif