BITS 32

%macro cglobal 1
  global _%1
  %define %1 _%1
%endmacro
%macro cextern 1
  extern _%1
  %define %1 _%1
%endmacro

;=============================================================================
; Read only data
;=============================================================================

%ifdef FORMAT_COFF
SECTION .rodata data
%else
SECTION .rodata data align=16
%endif

ALIGN 64
GetAbs: dd 0x7FFFFFFF, 0x7FFFFFFF
ALIGN 64
GetSign: dd 0x80000000, 0x80000000
ALIGN 64
MAD2sigma: dd 1.4826, 1.4826       ; =1/0.6745,    sigmaNoise = MAD /0.6745

;=============================================================================
; Code
;=============================================================================

SECTION .text

;cglobal  ht_3dne ; must be updated, so disabled in v.0.35
;cglobal  st_3dne ; must be updated, so disabled in v.0.35
;cglobal  qt_3dne ; must be updated, so disabled in v.0.35
cglobal  wiener_chop_3dne
cglobal  auto_th_3dne

;-----------------------------------------------------------------------------
; Helper macro HTONE
;-----------------------------------------------------------------------------
%macro HTONE 6
  movq    %2, [esi+eax+%1+8]
  movq    %3, mm7
  movq    %4, %2
  pandn   %5, %6
  pand    %2, mm3
  movq    [esi+eax+%1], %5
  pfcmpge %3, %2
%endmacro

%macro HTTWO 1
  HTONE   %1+0, mm1, mm5, mm4, mm6, mm2
  HTONE   %1+8, mm0, mm6, mm2, mm5, mm4
%endmacro
 
;-----------------------------------------------------------------------------
;
; Hard Thresholding
; ht_3dne(float* block, int b, float threshold);
;
;-----------------------------------------------------------------------------
ALIGN 16
ht_3dne:
  push  ebx
  push  esi
  mov   esi, [esp+8+ 4] ;block
  mov   ecx, [esp+8+ 8] ;b
  movd  mm7, [esp+8+12] ;threshold
  movq  mm3, [GetAbs]
  punpckldq mm7,mm7
  shl   ecx, 2
  mov   ebx, ecx
  xor   eax, eax
  and   ebx, -256

ALIGN 16
.loophard_aligned
  prefetchw [esi+eax+64] ;Unrolled loop - should be faster...
  movq    mm6, mm7
  movq    mm0, [esi+eax]
  movq    mm2, mm0
  pand    mm0, mm3
  pfcmpge mm6, mm0
    
  HTTWO    0
  HTTWO   16
  HTTWO   32
  HTONE   48, mm1, mm5, mm4, mm6, mm2

;  pfcmpge mm5, mm1 ; bug fixed: this line exist in macro and must be removed  from here - Fizick
  pandn   mm5, mm4
  movq    [esi+eax+56], mm5

  add     eax, 64
  cmp     eax, ebx
  jl      .loophard_aligned

  cmp     eax, ecx
  jnl     .loophard_end

ALIGN 16
.loophard_norm
  movq    mm0, [esi+eax]
  movq    mm2, mm0
  pand    mm0, mm3
  movq    mm6, mm7
  pfcmpge mm6, mm0
  pandn   mm6, mm2
  movq    [esi+eax], mm6
        
  add     eax,8
  cmp     eax,ecx
  jl      .loophard_norm

ALIGN 16
.loophard_end
  femms
  pop     esi
  pop     ebx
  ret

;-----------------------------------------------------------------------------
; Helper macro STONE
;-----------------------------------------------------------------------------
%macro STONE 1 ; %1 = offset
  movq    mm0,[esi+eax+%1]
  movq    mm2,mm0   ;mm2 = src
  movq    mm5,mm0   ;mm5 = src
  pand    mm0,mm3   ;mm0 = abs(src)
  pand    mm5,mm4   ;mm5 = sgn(src)
  movq    mm6,mm7   ;mm6 = thresh
  movq    mm1,mm7   ;mm1 = thresh
  pfcmpge mm0,mm6   ;mm0 = (abs(src)>=thresh)
  por     mm1,mm5   ;mm5 = sgn(src)*thresh
  pand    mm2,mm0   ;mm2 = (abs(src)>=thresh) ? src : 0
  pand    mm1,mm0   ;mm3 = (abs(src)>=thresh) ? sgn(src)*thresh : 0
  pfsub   mm2,mm1   ;mm3 = (abs(src)>=thresh) ? src + sgn(src)*thresh : 0
  movq    [esi+eax+%1],mm2
%endmacro

;-----------------------------------------------------------------------------
;
; Soft Thresholding
; st_3dne(float* block, int b, float threshold);
;
;-----------------------------------------------------------------------------
ALIGN 16
st_3dne:
  push  ebx
  push  esi
  mov   esi,[esp+8+ 4]
  mov   ecx,[esp+8+ 8]
;  movq  mm7,[esp+8+12] 
  movd  mm7,[esp+8+12] ; float threshold - Fizick
  punpckldq mm7,mm7
  shl   ecx,2
  mov   ebx,ecx
  movq  mm3,[GetAbs]
  xor   eax,eax
  movq  mm4,[GetSign]
  and   ebx,-256

ALIGN 16
.loopsoft_aligned 
  prefetchw [esi+eax+64] ;Unrolled loop - should be faster...
  STONE 0
  STONE 8
  STONE 16
  STONE 24
  STONE 32
  STONE 40
  STONE 48
  STONE 56

  add   eax,64
  cmp   eax,ecx
  jl    .loopsoft_aligned
 
  cmp   eax,ecx
  jnl   .loopsoft_end

ALIGN 16
.loopsoft_norm
  STONE 0

  add   eax,8
  cmp   eax,ecx
  jl    .loopsoft_norm

ALIGN 16
.loopsoft_end
  femms
  pop  esi
  pop  ebx
  ret

;-----------------------------------------------------------------------------
; Helper macro QTONE
;-----------------------------------------------------------------------------
%macro QTONE 1 ; %1 = offset
  movq      mm0,[esi+eax+%1]
  movq      mm5,mm0   ;mm5 = src1 | src2
  movq      mm2,mm0   ;mm2 = src1 | src2
  pfmul     mm5,mm5   ;mm5 = src1*src1 | src2*src2
  pand      mm0,mm3   ;mm0 = abs(src1)                           
  pfrcp     mm4,mm5   ;mm4 = 1/(src2*src2) 14bits precision
  pswapd    mm1,mm5   ;mm1 = src2*src2 | src1*src1
  pcmpgtd   mm0,mm7   ;mm0 = (abs(src)>=thresh)
  pfrcp     mm1,mm1   ;mm1 = 1/(src1*src1) 14bits precision
;  pfsub     mm5,mm7   ;mm5 = src*src - thresh*thresh   ; bug: mm7=thresh 
  pfsub     mm5,mm6   ;mm5 = src*src - thresh*thresh   ; mm6=thresh*thresh   - Fizick
  pfmul     mm2,mm5   ;mm2 = src * (src*src - thresh*thresh)
  punpckldq mm4,mm1   ;mm1 = 1/(src1*src1) | 1/(src2*src2)
  pfmul     mm2,mm4   ;mm2 = src * (src*src - thresh*thresh) * 1 / (src*src)
  pand      mm2,mm0   ;mm2 = (abs(src)>=thresh) ?
                      ;      src * (src*src - thresh*thresh) / (src*src) : 0
  movq  [esi+eax+%1],mm2
%endmacro

;-----------------------------------------------------------------------------
;
; Qian Thresholding
; qt_3dne(float* block, int b, float threshold);
;
;-----------------------------------------------------------------------------
ALIGN 16
qt_3dne:
  push  ebx
  push  esi
  mov   esi,[esp+8+ 4]
  mov   ecx,[esp+8+ 8]
;  movq  mm7,[esp+8+12];  
  movd  mm7,[esp+8+12] ; threshold is float dword - Fizick
  punpckldq mm7,mm7
  shl   ecx,2 ; get length in bytes
  mov   ebx,ecx
  movq  mm3,[GetAbs]
  xor   eax,eax
  movq  mm4,[GetSign]
;  and   ebx,-255 
  and   ebx,-256 ; probably must be -256=FFFFFF00, and where it is used? - Fizick
  movq mm6,mm7  ; mm6 = thresh          - Fizick
  pfmul mm6,mm6 ; mm6=thresh*thresh     (as in reference  VD_3DNow.cpp code by Kurosu) - Fizick
  
ALIGN 16
.loopqian_aligned 
  prefetchw [esi+eax+64] ;Unrolled loop - should be faster...
  QTONE  0
  QTONE  8
  QTONE 16
  QTONE 24
  QTONE 32
  QTONE 40
  QTONE 48
  QTONE 56

  add   eax,64
;  cmp   eax,ecx 
  cmp   eax,ebx ; probably must be ebx (mod64) here - Fizick
  jl    .loopqian_aligned
 
  cmp   eax,ecx
  jnl   .loopqian_end

ALIGN 16
.loopqian_norm
  QTONE 0
  add   eax,8	; therefore, ecx must be mod8,
				; and length b=plane_width*plane_height is assumed even,
				; for YV12 color planes also - Fizick
  cmp   eax,ecx
  jl    .loopqian_norm

ALIGN 16
.loopqian_end
  femms
  pop   esi
  pop   ebx
  ret


;-----------------------------------------------------------------------------
;
; wiener_chop_3dne(const float* block1, float *block,
;                  const uint32_t length, float sigmaNoise)
;
;-----------------------------------------------------------------------------
ALIGN 16
wiener_chop_3dne:
  ;push    ebx
  push    esi
  push    edi
  movd    mm7, [esp+12+12]	; mm7 = 0 | sigmaNoise
  mov     esi, [esp+12+ 0]	; block1
  punpckldq mm7, mm7		; mm7 = sigmaNoise | sigmaNoise
  mov     edi, [esp+12+ 4]	; block
  pfmul   mm7, mm7		; mm7 = varNoise | varNoise
  mov     ecx, [esp+12+ 8]	; length
  xor     eax, eax		; iterator
  shl     ecx, 2		; ecx = length*sizeof(float)

;; Could be much much unrolled to hide pfmul/pfrcp latency
ALIGN16
.loopchop:
  movq    mm0, [esi+eax]	; mm0 = b1[0] | b1[1]
  movq    mm1, [edi+eax]	; mm1 = b[0] | b[1]
  pfmul   mm0, mm0		; mm0 = b1 = varest
  movq    mm2, mm0		; mm2 = varest
  pfadd   mm2, mm7		; mm2 = varest+varnoise = v
  pfrcp   mm3, mm2		; mm3 = 1/v[1] 14bits precision
  pfmul   mm1, mm0		; mm1 = b*varest
  pswapd  mm2, mm2		; mm0 = v[1] | v[0]
  pfrcp   mm2, mm2		; mm0 = 1/v[0] 14bits precision
  punpckldq mm3, mm2		; mm3 = 1/v[0] | 1/v[1]
  pfmul   mm1, mm3		; mm1 = b*varest/v
  movq    [edi+eax], mm1	; b = b*varest/v

  add     eax, 8
  cmp	  eax, ecx
  jl      .loopchop

  emms
  pop     edi
  pop     esi
  ;pop     ebx
  ret

;-----------------------------------------------------------------------------
;
; auto_th_3dne(const float *block, int w, int h, int nsteps, float ratio,
;              float * thresholdEstimated
; Should be unrolled as it's a lot of latency in fact
; modified in v0.34.3
;-----------------------------------------------------------------------------
ALIGN 16
auto_th_3dne:
  push    ebx
  push    esi
  push edi
  mov     esi, [esp+16+ 0]	; block
  mov     eax, [esp+16+ 4]	; eax = w
  mov	  ebx, [esp+16 +8] ; ebx = h;
  shr	ebx, 1; h/2
  mov edi, ebx;  h/2
  mov		ecx, eax;
  shl	 ecx, 2;		w*4;
  shr	eax, 1; w/2
  imul    eax, ebx	; eax = low32(w/2*h/2)
  mov     edx, eax		; edx = length4=w*h/4    number of values
  movd	mm4, edx;
	punpckldq mm4, mm4 ; 
  pi2fd mm4, mm4;  w*h/4    number of values length4
  pfrcp   mm4, mm4		; mm4 = 1/float(length4)
  shl	eax, 3; w*h/4*4*2
  add esi, eax;
  mov edx, ecx;
  shr edx, 1; w/2 *4
  add esi, edx; 

   pxor	  mm0, mm0		; mm0 = 0
  movq    mm3, [GetAbs]		; mm3 = 0x7FFFFFFF | 0x7FFFFFFF
  xor     ebx, ebx		; ebx = 0     h counter
  xor     eax, eax		; eax = 0     w counter

ALIGN 16
.loop_auto:
  movq    mm1, [esi+eax]	; mm1 = block
  pand    mm1, mm3		; mm1 = fabs(block)
  pfadd   mm0, mm1		; mm0 += fabs(block)

  add     eax, 8
  cmp     eax, edx ; w/2 *4
  jl      .loop_auto
	xor eax, eax		; eax = 0
	add esi, ecx; w*4
	add ebx, 1;
	cmp ebx, edi;  h/2
	 jl .loop_auto

						; fixed and changed by Fizick
  pfmul		mm0, mm4		; mm0 = sum MAD * 1/float(length4) 
  pfacc   mm0, mm0			; mm0 = MAD[0] + MAD[1]
  pfmul   mm0, [MAD2sigma]	; mm0 = MAD  /0.6745 
  pfrcp mm5, [esp+16+16]		; 1/ratio
  pfmul mm0, mm5			; mm0 = thresh = MAD /0.6745 /ratio
  mov	esi, [esp+16+20]	
  movd	  [esi], mm0:		; return thresh  
  emms
  pop edi
  pop     esi
  pop     ebx
  ret
  