BITS 32

%macro cglobal 1
  global _%1
  %define %1 _%1
%endmacro
%macro cextern 1
  extern _%1
  %define %1 _%1
%endmacro

;=============================================================================
; Read only data
;=============================================================================

%ifdef FORMAT_COFF
SECTION .rodata data
%else
SECTION .rodata data align=16
%endif

ALIGN 64
GetAbs:  dd 0x7FFFFFFF, 0x7FFFFFFF, 0x7FFFFFFF, 0x7FFFFFFF
GetSign: dd 0x80000000, 0x80000000, 0x80000000, 0x80000000
f01: dd 0.01, 0.01, 0.01, 0.01
f1:  dd 1.0, 1.0, 1.0, 1.0
sfactor: dd 2.0, 2.0, 2.0, 2.0

;=============================================================================
; Code
;=============================================================================

SECTION .text

cglobal  ht_sse
;cglobal  st_sse; must be updated, so disabled in v.0.35
cglobal  qt_sse
cglobal  smt_sse

;-----------------------------------------------------------------------------
;
; Hard Thresholding
; ht_sse(float* block, int b, float threshold,
;		   int width,  int height,  int nsteps, float percent) - updated in v.035
;
;-----------------------------------------------------------------------------
%macro HTONE 1 ; %1 = offset  
movaps  xmm0,[esi+eax +%1]
  movaps  xmm1,xmm0 ; b
  andps   xmm0,xmm3 ; getabs
  cmpltps xmm0,xmm7 ; cmp b, threshold
  movaps xmm4, xmm0;
  andnps  xmm4,xmm1 ; b if not condition
	mulps xmm1, xmm2; b*frac
	andps xmm0, xmm1; b*frac if condition
	orps xmm0, xmm4;
  movaps  [esi+eax +%1],xmm0
%endmacro
;-----------------------------------------------------------------------------

ALIGN 16
ht_sse:
  push    ebx
  push    esi
  mov     esi,[esp+8+ 4]
  mov     ecx,[esp+8+ 8]
  movss   xmm7,[esp+8+12]
  movaps  xmm3,[GetAbs]
  shufps  xmm7,xmm7,0x0000
  movss   xmm2,[esp+8+28] ; percent
  shufps  xmm2,xmm2,0x0
  movaps xmm0, [f01] ; 0.1
  mulps	xmm0, xmm2 ; percent*0.01
  movaps xmm2, [f1] ; 1.0
  subps	xmm2, xmm0 ; frac=1-p01
  ;mov     ebx,ecx
  ;xor     eax,eax
  ;and     ebx,-15
  ;shl     ecx,2
  ;shl     ebx,2
  shl   ecx,2
  mov   ebx,ecx
  xor   eax,eax
  and   ebx,-256

ALIGN 16
.loophard_aligned:
  prefetchnta [esi+eax+64]
  HTONE    0
  HTONE   16
  HTONE   32
  HTONE   48
   
  add     eax,64
  cmp     eax,ebx
  jl      .loophard_aligned
  cmp     eax,ecx
  jl      .loophard_norm
  jmp     loophard_end

ALIGN 16
.loophard_norm
  HTONE    0
    
  add     eax,16
  cmp     eax,ecx
  jl      .loophard_norm

ALIGN 16
loophard_end:
  pop     esi
  pop     ebx
  emms ; added by Fizick in v.0.32.1
  ret

;-----------------------------------------------------------------------------
; Helper macro STONE
;-----------------------------------------------------------------------------
%macro STONE 1 ; %1 = offset
  movaps  xmm0,[esi+eax+%1]
  movaps  xmm2,xmm0
  movaps  xmm5,xmm0
  andps   xmm0,xmm3
  andps   xmm5,xmm4
  movaps  xmm6,xmm7
  movaps  xmm1,xmm7
  cmpltps xmm6,xmm0
  orps    xmm1,xmm5
  subps   xmm2,xmm1
  andps   xmm2,xmm6
  movaps  [esi+eax+%1],xmm2
%endmacro

;-----------------------------------------------------------------------------
;
; Soft Thresholding - obsolete, disabled, must be updated to v0.35
; st_sse(float* block, int b, float threshold);
;
;-----------------------------------------------------------------------------
ALIGN 16
st_sse:
  push    ebx
  push    esi
  mov     esi,[esp+8+ 4]
  mov     ecx,[esp+8+ 8]
  movss   xmm7,[esp+8+12]
  movaps  xmm3,[GetAbs]
  movaps  xmm4,[GetSign]
  shufps  xmm7,xmm7,0x0000
  mov     ebx,ecx
  xor     eax,eax
  and     ebx,-15
  shl     ecx,2
  shl     ebx,2

ALIGN 16
.loopsoft_aligned
  prefetchnta [esi+eax+128]
  STONE    0
  STONE   16
  STONE   32
  STONE   48

  add     eax,64
  cmp     eax,ebx
  jl      .loopsoft_aligned
   
  cmp     eax,ecx
  jnl     .loopsoft_end

ALIGN 16
.loopsoft_norm
  STONE   0
  add     eax,16
  cmp     eax,ecx
  jl      .loopsoft_norm

ALIGN 16
.loopsoft_end
  pop     esi
  pop     ebx
  emms ; added in v.0.32.1 by Fizick
  ret

;-----------------------------------------------------------------------------
; Helper macro QTONE - used in old versions before 0.35
;-----------------------------------------------------------------------------
%macro QTONE 1 ; %1 = offset
  movaps  xmm0,[esi+eax+%1]
  movaps  xmm5,xmm0
  movaps  xmm2,xmm0
  mulps   xmm5,xmm5
  andps   xmm0,xmm3
  movaps  xmm1,xmm7
  rcpps   xmm4,xmm5
  subps   xmm5,xmm6
  mulps   xmm5,xmm2
  cmpltps xmm1,xmm0
  mulps   xmm5,xmm4
  andps   xmm5,xmm1
  movaps  [esi+eax+%1],xmm5
%endmacro

;-----------------------------------------------------------------------------
; Helper macro QTONEP - partial denoise  - updated v.035 by Fizick
; input: xmm3=GetAbs, xmm6=thresh*thresh, xmm7=thresh, xmm2=percent01
;-----------------------------------------------------------------------------
%macro QTONEP 1 ; %1 = offset
  movaps  xmm0,[esi+eax+%1]
  movaps  xmm5,xmm0 ; b
  mulps   xmm5,xmm5 ;b*b
  movaps  xmm1,xmm7 ; t
  rcpps   xmm4,xmm5 ; 1/b*b
	movaps xmm0, xmm2 ; per	
	mulps xmm0, xmm6;  tr2=t*t*per
  subps   xmm5,xmm0  ; b*b - tr2
	movaps  xmm0,[esi+eax+%1] ; b again
  mulps   xmm5,xmm0 ; (b*b - tr2)*b 
  mulps   xmm5,xmm4 ; (b*b - tr2)*b/(b*b)
	movaps xmm4, xmm0;  b
	andps   xmm4,xmm3 ; abs(b)
  cmpltps xmm1,xmm4 ; cmp t, abs(b)
  andps   xmm5,xmm1 ; this values of condition true
	movaps xmm4, xmm0; b
	mulps xmm4, xmm2; b*percent01
	subps xmm0, xmm4; b-b*percent01
	andnps xmm1, xmm0 ; get other branch
	orps xmm5, xmm1
  movaps  [esi+eax+%1],xmm5
%endmacro

;-----------------------------------------------------------------------------
;
; Qian Thresholding 
; qt_sse(float* block, int len, float threshold,
;		   int width,  int height,  int nsteps, float percent) - updated in v.035
;
;-----------------------------------------------------------------------------
ALIGN 16
qt_sse:
  push    ebx
  push    esi
  mov     esi,[esp+8+ 4] ;block
  mov     ecx,[esp+8+ 8] ;len
  movss   xmm7,[esp+8+12] ; thresh
  shufps  xmm7,xmm7,0x0
  movss   xmm2,[esp+8+28] ; percent
  shufps  xmm2,xmm2,0x0
  movaps xmm0, [f01]
  mulps	xmm2, xmm0 ; percent*0.01
  movaps  xmm3,[GetAbs]
  mov     ebx,ecx
  xor     eax,eax
  movaps  xmm6,xmm7
  and     ebx,-15
  mulps	  xmm6,xmm7 ; thresh*thresh
  shl     ecx,2
  shl     ebx,2

ALIGN 16
.loopqian_aligned
  prefetchnta [esi+eax+64]
  QTONEP    0
  QTONEP   16
  QTONEP   32
  QTONEP   48

  add     eax,64
  cmp     eax,ebx
  jl      .loopqian_aligned
   
  cmp     eax,ecx
  jl      .loopqian_norm
  jmp     .loopqian_end

ALIGN 16
.loopqian_norm
  QTONEP   0
  add     eax,16
  cmp     eax,ecx
  jl      .loopqian_norm

ALIGN 16
.loopqian_end
  pop     esi
  pop     ebx
  emms ; added in v.0.32.1 by Fizick
  ret

;-----------------------------------------------------------------------------
; Helper macro SMTONE -  v.035 by Fizick
; input: xmm6=thresh*thresh, xmm7=tr2frac
;-----------------------------------------------------------------------------
;		float temp = block[i];
;		float tp2 = temp*temp;
;		block[i] = temp*(tp2+tr2frac)/(tp2 + tr2);
;----------------------------------------------------
%macro SMTONE 1 ; %1 = offset
  movaps  xmm0,[esi+eax+%1]
  movaps  xmm5,xmm0 ; temp=b
  mulps   xmm5,xmm5 ; tp2=b*b
  movaps  xmm1,xmm6 ; tr2
  addps	xmm1, xmm5; tr2+tp2
  addps xmm5, xmm7; tp2 + tr2frac 
  rcpps   xmm1,xmm1 ; 1/(tp2+tr2)
  mulps	xmm5, xmm1; (tp2+tr2frac)/(tp2 + tr2)
  mulps xmm0, xmm5; b*(tp2+tr2frac)/(tp2 + tr2)
	movaps  [esi+eax+%1],xmm0
%endmacro

;-----------------------------------------------------------------------------
;
; Smooth Thresholding - added in v.035
; smootht_sse(float* block, int len, float threshold,
;		   int width,  int height,  int nsteps, float percent) 
;
;-----------------------------------------------------------------------------
ALIGN 16
smt_sse:
  push    ebx
  push    esi
  mov     esi,[esp+8+ 4] ;block
  mov     ecx,[esp+8+ 8] ;len
  movss   xmm6,[esp+8+12] ; thresh
  shufps  xmm6,xmm6,0x0
  mulps	  xmm6,xmm6 ; thresh*thresh
  movaps xmm1, [sfactor]; 2.0
  shufps  xmm1,xmm1,0x0
  mulps xmm6, xmm1; thresh*thresh*sfactor
  movss   xmm1,[esp+8+28] ; percent
  shufps  xmm1,xmm1,0x0
  movaps xmm0, [f01]; 0.01
  mulps	xmm1, xmm0 ; percent*0.01
  movaps  xmm7,[f1] ; 1.0
  subps	xmm7, xmm1; 1-percent*0.01
	mulps xmm7,xmm6 ; (1-percent*0.01)*thresh*thresh
  mov     ebx,ecx
  xor     eax,eax
  and     ebx,-15
  shl     ecx,2
  shl     ebx,2

ALIGN 16
.loopsmt_aligned
  prefetchnta [esi+eax+64]
  SMTONE    0
  SMTONE   16
  SMTONE   32
  SMTONE   48

  add     eax,64
  cmp     eax,ebx
  jl      .loopsmt_aligned
   
  cmp     eax,ecx
  jl      .loopsmt_norm
  jmp     .loopsmt_end

ALIGN 16
.loopsmt_norm
  SMTONE   0
  add     eax,16
  cmp     eax,ecx
  jl      .loopsmt_norm

ALIGN 16
.loopsmt_end
  pop     esi
  pop     ebx
  emms 
  ret