;/*******************
; * Taken from http://www.john.findlay1.btinternet.co.uk/asm/memset.htm
; *******************/
; Adapted for Vaguedenoiser (inline asm) by Kurosu
; ported to NASM by Fizick in v.0.32
;
BITS 32

%macro cglobal 1
  global _%1
  %define %1 _%1
%endmacro
%macro cextern 1
  extern _%1
  %define %1 _%1
%endmacro


%define USE_MOVNTQ2		0

%if USE_MOVNTQ2
%define MOVQ2  movntq
%else
%define MOVQ2  movq
%endif

;=======================================================================
;__inline void Wavelet::mem_set(void * dest, int set, size_t nBytes)
;{
;	__asm
;	{
;=======================================================================

SECTION .text

cglobal  mem_set

;-------------------------------------------------
mem_set:
ALIGN 16
		push		ebx

		push edi
			
		mov			edi,[esp+8+4]	; dest
		mov			ebx,[esp+8+8]	; set 
		mov			edx,[esp+8+12]	; nBytes
		mov			eax,edx
			
		;// if num bytes is very small go straight to the last section
		;// this will also handle zero num bytes
		cmp			eax,16
;		jl			L8check ; too long jump for debug - replaced to two jumps in v.0.33.1 
		jnl			noL8check
		jmp			L8check

noL8check:			
		movd		mm0,[esp+8+8]		; set all eight bytes from memory into mm0
		punpcklbw	mm0,mm0
		
		;// align src on 8 boundary. If memory is non-aligned it will seriously
		;// affect performance, one cannot align both src & dest
		mov			ecx,edi
		punpcklwd	mm0,mm0
		punpckldq	mm0,mm0
		and			ecx,7
		jecxz		L64check			;// jump if no parity
		mov			eax,8
		sub			eax,ecx
		mov			ecx,eax
aligned:
		mov			di,bx
		inc			edi
		dec			ecx
		;//cmp			ecx,0
		jnz			aligned
			
L64check:
		mov			ecx,edx
		shr			ecx,6
		mov			eax,ecx				;// tmp
		jecxz		L32check
		shl			eax,6
		sub			edx,eax		;// new value of num bytes (orig - 64's)

		;//align		16	
L64:
		jecxz		L32check			;// jump to next section if finished
		dec			ecx
;		prefetchw  [edi+64] ; // changed from prefetcht0 for compatibility with old CPU in v.0.34.1 and finally disabled
		MOVQ2		[edi+0],mm0
		MOVQ2		[edi+8],mm0
		MOVQ2		[edi+16],mm0
		MOVQ2		[edi+24],mm0
		MOVQ2		[edi+32],mm0
		MOVQ2		[edi+40],mm0
		MOVQ2		[edi+48],mm0
		MOVQ2		[edi+56],mm0
		add			edi,64					;// add 64 each loop
%if USE_MOVNTQ2
		sfence
%endif
		jmp			L64
			
L32check:
		mov			ecx,edx
		shr			ecx,5
		mov			eax,ecx
		jecxz		L16check
		shl			eax,5
		sub			edx,eax
			
		;//align		16	
;//L32:									;// can only be one 32's to copy, no loop
		dec			ecx
		movq		[edi+0],mm0
		movq		[edi+8],mm0
		movq		[edi+16],mm0
		movq		[edi+24],mm0
		add			edi,32
			
L16check:
		mov			ecx,edx
		shr			ecx,4
		mov			eax,ecx
		jecxz		L8check
		shl			eax,4
		sub			edx,eax
			
		;//align		16	
;//L16:									;// can only be one 16's to copy
		MOVQ		[edi+0],mm0
		MOVQ		[edi+8],mm0
		add			edi,16
			
L8check:
		mov			ecx,edx
		shr			ecx,3
		mov			eax,ecx
		jecxz		L8
		shl			eax,3
		sub			edx,eax

		MOVQ		[edi+0],mm0			;// can only be one 32's to copy, no loop
		add			edi,8
								

L8:
		mov			ecx,edx
Lcheck:									;// copy the last remaining bytes
		jecxz		end1
		mov			di,bx
		inc			edi
		dec			ecx
		;//cmp			ecx,0
		jnz			Lcheck
			
end1:
		
		pop edi

		pop			ebx
		emms
		ret
;	}
;}
;%endif
